from udsoncan.Response import Response
from udsoncan.exceptions import *
import inspect
import sys
from abc import ABC

def cls_from_request_id(given_id):
	return BaseService.from_request_id(given_id)

def cls_from_response_id(given_id):
	return BaseService.from_response_id(given_id)

class BaseSubfunction:
	
	@classmethod
	def get_name(cls, subfn_id):
		attributes = inspect.getmembers(cls, lambda a:not(inspect.isroutine(a)))
		subfn_list = [a for a in attributes if not(a[0].startswith('__') and a[0].endswith('__'))]

		for subfn in subfn_list:
			if isinstance(subfn[1], int):
				if subfn[1] == subfn_id:	# [1] is value
					return subfn[0] 		# [0] is property name
			elif isinstance(subfn[1], tuple):
				if subfn_id >= subfn[1][0] or subfn_id <= subfn[1][1]:
					return subfn[0] 
		name = cls.__name__ if not hasattr(cls, '__pretty_name__') else cls.__pretty_name__
		return 'custom %s' % name

class BaseService(ABC):

	always_valid_negative_response = [
		Response.Code.GeneralReject,
		Response.Code.ServiceNotSupported,
		Response.Code.ResponseTooLong,
		Response.Code.BusyRepeatRequest,
		Response.Code.NoResponseFromSubnetComponent,
		Response.Code.FailurePreventsExecutionOfRequestedAction,
		Response.Code.SecurityAccessDenied, # ISO-14229:2006 Table A.1:  "Besides the mandatory use of this negative response code as specified in the applicable services within ISO 14229, this negative response code can also be used for any case where security is required and is not yet granted to perform the required service."
		Response.Code.RequestCorrectlyReceived_ResponsePending,
		Response.Code.ServiceNotSupportedInActiveSession
	]

	@classmethod	# Returns the service ID used for a client request
	def request_id(cls):
		return cls._sid

	@classmethod	# Returns the service ID used for a server response
	def response_id(cls):
		return cls._sid + 0x40

	@classmethod	# Returns an instance of the service identified by the service ID (Request)
	def from_request_id(cls, given_id):
		for name, obj in inspect.getmembers(sys.modules[__name__]):
			if hasattr(obj, "__bases__") and cls in obj.__bases__:
				if obj.request_id() == given_id:
					return obj

	@classmethod	# Returns an instance of the service identified by the service ID (Response)
	def from_response_id(cls, given_id):

		for name, obj in inspect.getmembers(sys.modules[__name__]):
			if hasattr(obj, "__bases__") and cls in obj.__bases__:
				if obj.response_id() == int(given_id):
					return obj

	#Default subfunction ID for service that does not implement subfunction_id().
	def subfunction_id(self):
		return 0

	@classmethod	# Tells if this service includes a subfunction byte
	def use_subfunction(cls):
		if hasattr(cls, '_use_subfunction'):
			return cls._use_subfunction
		else:
			return True
	@classmethod
	def has_response_data(cls):
		if hasattr(cls, '_no_response_data'):
			return False if cls._no_response_data else True
		else:
			return True

	@classmethod	# Returns the service name. Shortcut that works on class and instances
	def get_name(cls):
		return cls.__name__

	@classmethod	# Tells if the given response code is expected for this service according to UDS standard.
	def is_supported_negative_response(cls, code):
		supported = False
		if code in cls.supported_negative_response:
			supported = True

		if code in cls.always_valid_negative_response:
			supported = True
		
		# As specified by Annex A, negative response code ranging above 0x7F can be used anytime if the service can return ConditionNotCorrect
		if code >= 0x80 and code < 0xFF and Response.Code.ConditionsNotCorrect in cls.supported_negative_response:
			supported = True

		# ISO-14229:2006 Table A.1 : "This response code shall be supported by each diagnostic service with a subfunction parameter"
		if code == Response.Code.SubFunctionNotSupportedInActiveSession and cls.use_subfunction():
			supported = True

		return supported

def is_valid_service(service_cls):
	return issubclass(service_cls, BaseService)

class ServiceHelper:

	@staticmethod
	def validate_int(value, min=0, max=0xFF, name='value'):
		if not isinstance(value, int):
			raise ValueError("%s must be a valid integer" % (name))
		if value < min or value > max:
			raise ValueError("%s   must be an integer between 0x%X and 0x%X" % (name, min, max))

	# Make sure that the actual client configuration contains valid definitions for given Data Identifiers
	@staticmethod
	def check_did_config(didlist, didconfig):
		didlist = [didlist] if not isinstance(didlist, list) else didlist
		if 'data_identifiers' in didconfig:
			didconfig = config['data_identifiers']

		for did in didlist:
			if did not in didconfig:
				raise ConfigError(did, msg='Actual data identifier configuration contains no definition for data identifier 0x%04x' % did)
	
		return didconfig

	# Make sure that the actual client configuration contains valid definitions for given Input/Output Data Identifiers
	@staticmethod
	def check_io_config(didlist, ioconfig):
		didlist = [didlist] if not isinstance(didlist, list) else didlist
		if 'input_output' in ioconfig:
			ioconfig = ioconfig['input_output']

		if not isinstance(ioconfig, dict):
			raise ConfigError('input_output', msg='Configuration of Input/Output section must be a dict.')

		for did in didlist:
			if did not in ioconfig:
				raise ConfigError(key=did, msg='Actual Input/Output configuration contains no definition for data identifier 0x%04x' % did)
			if isinstance(ioconfig[did], dict):	# IO Control services has that concept of composite DID. We define them with dicts.
				if 'codec'not in ioconfig[did]:
					raise ConfigError('codec', msg='Configuration for Input/Output identifier 0x%04x is missing a codec')

				if 'mask' in ioconfig[did]:
					mask_def = ioconfig[did]['mask']
					for mask_name in mask_def:
						if not isinstance(mask_def[mask_name], int):
							raise ValueError('In Input/Output configuration for did 0x%04x, mask "%s" is not an integer' % (did, mask_name))

						if mask_def[mask_name] < 0:
							raise ValueError('In Input/Output configuration for did 0x%04x, mask "%s" is not a positive integer' % (did, mask_name))

				
				if 'mask_size' in ioconfig[did]:
					if not isinstance(ioconfig[did]['mask_size'], int):
						raise ValueError('mask_size in Input/Output configuration for did 0x%04x must be a valid integer' % (did))

					if ioconfig[did]['mask_size'] < 0:
						raise ValueError('mask_size in Input/Output configuration for did 0x%04x must be greater than 0' % (did))

					if 'mask' in ioconfig[did]:
						mask_def = ioconfig[did]['mask']
						for mask_name in mask_def:
							if mask_def[mask_name] > 2**(ioconfig[did]['mask_size']*8)-1:
								raise ValueError('In Input/Output configuration for did 0x%04x, mask "%s" cannot fit in %d bytes (defined by mask_size)' % (did, mask_name,ioconfig[did]['mask_size']))

		return ioconfig

class BaseResponseData:
	def __init__(self, service_class):
		if not issubclass(service_class, BaseService):
			raise ValueError('service_class must be a service class')

		self.service_class = service_class

	def __repr__(self):
		return '<%s (%s) at 0x%08x>' % (self.__class__.__name__, self.service_class.__name__, id(self))

from .DiagnosticSessionControl import DiagnosticSessionControl
from .ECUReset import ECUReset
from .SecurityAccess import SecurityAccess
from .CommunicationControl import CommunicationControl
from .AccessTimingParameter import AccessTimingParameter
from .SecuredDataTransmission import SecuredDataTransmission
from .TesterPresent import TesterPresent
from .ControlDTCSetting import ControlDTCSetting
from .ResponseOnEvent import ResponseOnEvent
from .LinkControl import LinkControl
from .ReadDataByIdentifier import ReadDataByIdentifier
from .WriteDataByIdentifier import WriteDataByIdentifier
from .ReadMemoryByAddress import ReadMemoryByAddress
from .InputOutputControlByIdentifier import InputOutputControlByIdentifier
from .RoutineControl import RoutineControl
from .ReadScalingDataByIdentifier import ReadScalingDataByIdentifier
from .ReadDataByPeriodicIdentifier import ReadDataByPeriodicIdentifier
from .WriteMemoryByAddress import WriteMemoryByAddress
from .DynamicallyDefineDataIdentifier import DynamicallyDefineDataIdentifier
from .ClearDiagnosticInformation import ClearDiagnosticInformation
from .ReadDTCInformation import ReadDTCInformation
from .RequestDownload import RequestDownload
from .RequestUpload import RequestUpload
from .TransferData import TransferData
from .RequestTransferExit import RequestTransferExit



from udsoncan.client import Client
from udsoncan import services
from udsoncan.exceptions import *
from udsoncan import DidCodec
from udsoncan import IOValues, IOMasks
import struct

from test.ClientServerTest import ClientServerTest

class StubbedCodec(DidCodec):
	def encode(self, did_value):
		return struct.pack('B', did_value+1)

	def decode(self, did_payload):
		return struct.unpack('B', did_payload)[0] - 1

	def __len__(self):
		return 1

class StubbedCompositeCodec(DidCodec):
	def encode(self, IAC_pintle, rpm, pedalA, pedalB, EGR_duty):
		pedal = (pedalA << 4) | pedalB
		return struct.pack('>BHBB', IAC_pintle, rpm, pedal, EGR_duty)

	def decode(self, payload):
		vals = struct.unpack('>BHBB', payload)
		return {
			'IAC_pintle': vals[0],
			'rpm' 		: vals[1],
			'pedalA' 	: (vals[2] >> 4) & 0xF,
			'pedalB' 	: vals[2] & 0xF,
			'EGR_duty' 	: vals[3]
		}

	def __len__(self):
		return 5		

class TestIOControl(ClientServerTest):
	def __init__(self, *args, **kwargs):
		ClientServerTest.__init__(self, *args, **kwargs)

	def postClientSetUp(self):
		self.udsclient.config["input_output"] = {
			0x132 : StubbedCodec,
			0x456 : '<HH',
			0x155 : {
				'codec' : StubbedCompositeCodec,
				'mask' : {
					'IAC_pintle': 0x80,
					'rpm' 		: 0x40,
					'pedalA' 	: 0x20,
					'pedalB' 	: 0x10,
					'EGR_duty'	: 0x08
				},
				'mask_size' : 2
			}
		}

#As defined by ISO-14229:2006, section 12.2.5.2 (Example #1)
	def test_io_control_single_reset(self):
		request = self.conn.touserqueue.get(timeout=0.2)
		self.assertEqual(request, b"\x2F\x01\x32\x01")
		self.conn.fromuserqueue.put(b"\x6F\x01\x32\x01\x4B")	# Positive response

	def _test_io_control_single_reset(self):
		response = self.udsclient.io_control(control_param=1, did=0x132)	# Reset to default
		self.assertEqual(response.service_data.control_param_echo, 1)	
		self.assertEqual(response.service_data.did_echo, 0x132)	
		self.assertEqual(response.service_data.decoded_data, 0x4A)	# 0x4B-1 as defined by codec decode method

	def test_io_control_single_reset_spr_no_effect(self):
		request = self.conn.touserqueue.get(timeout=0.2)
		self.assertEqual(request, b"\x2F\x01\x32\x01")
		self.conn.fromuserqueue.put(b"\x6F\x01\x32\x01\x4B")	# Positive response

	def _test_io_control_single_reset_spr_no_effect(self):
		with self.udsclient.suppress_positive_response:
			response = self.udsclient.io_control(control_param=1, did=0x132)	# Reset to default
			self.assertEqual(response.service_data.control_param_echo, 1)	
			self.assertEqual(response.service_data.did_echo, 0x132)	
			self.assertEqual(response.service_data.decoded_data, 0x4A)	# 0x4B-1 as defined by codec decode method

	def test_io_control_no_control_param(self):
		request = self.conn.touserqueue.get(timeout=0.2)
		self.assertEqual(request, b"\x2F\x01\x32\x78")
		self.conn.fromuserqueue.put(b"\x6F\x01\x32\x4B")

	def _test_io_control_no_control_param(self):
		response = self.udsclient.io_control(did=0x132, values=[0x77]) # No control_param, skip directly to data	
		self.assertEqual(response.service_data.control_param_echo, None)	
		self.assertEqual(response.service_data.did_echo, 0x132)	
		self.assertEqual(response.service_data.decoded_data, 0x4A)	# 0x4B-1 as defined by codec decode method

	def test_io_control_with_repsonse_record(self):
		request = self.conn.touserqueue.get(timeout=0.2)
		self.assertEqual(request, b"\x2F\x04\x56\x03\x11\x01\x22\x02")
		self.conn.fromuserqueue.put(b"\x6F\x04\x56\x03\x33\x03\x44\x04")	# Positive response with 0x333 and 0x444 as response data

	def _test_io_control_with_repsonse_record(self):
		response = self.udsclient.io_control(control_param=3, did=0x456, values=IOValues(0x111,0x222))	# Short Term Adjustment
		self.assertEqual(response.service_data.decoded_data, (0x333, 0x444))	

	def test_io_control_with_repsonse_record_zero_padding_tolerated(self):
		data = b'\x6F\x04\x56\x03\x33\x03\x44\x04'

		for i in range(3):
			self.wait_request_and_respond(data + b'\x00' * (i+1))
		
	def _test_io_control_with_repsonse_record_zero_padding_tolerated(self):
		self.udsclient.config['tolerate_zero_padding'] = True

		for i in range(3):
			response = self.udsclient.io_control(control_param=3, did=0x456, values=IOValues(0x111,0x222))	
			self.assertEqual(response.service_data.control_param_echo, 3)
			self.assertEqual(response.service_data.did_echo, 0x456)
			self.assertEqual(response.service_data.decoded_data, (0x333, 0x444))

	def test_io_control_with_repsonse_record_zero_padding_not_tolerated_exception(self):
		data = b'\x6F\x04\x56\x03\x33\x03\x44\x04'
		for i in range(3):
			self.wait_request_and_respond(data + b'\x00' * (i+1))

	def _test_io_control_with_repsonse_record_zero_padding_not_tolerated_exception(self):
		self.udsclient.config['tolerate_zero_padding'] = False
		for i in range(3):
			with self.assertRaises(InvalidResponseException):
				self.udsclient.io_control(control_param=3, did=0x456, values=IOValues(0x111,0x222))	

	def test_io_control_with_repsonse_record_zero_padding_not_tolerated_no_exception(self):
		data = b'\x6F\x04\x56\x03\x33\x03\x44\x04'
		for i in range(3):
			self.wait_request_and_respond(data + b'\x00' * (i+1))

	def _test_io_control_with_repsonse_record_zero_padding_not_tolerated_no_exception(self):
		self.udsclient.config['exception_on_invalid_response'] = False
		self.udsclient.config['tolerate_zero_padding'] = False
		
		for i in range(3):
			response = self.udsclient.io_control(control_param=3, did=0x456, values=IOValues(0x111,0x222))	
			self.assertFalse(response.valid)

	def test_io_control_composite_did_dict(self):
		request = self.conn.touserqueue.get(timeout=0.2)
		self.assertEqual(request, b"\x2F\x01\x55\x03\x07\x12\x34\x45\x99\x00\xA0")
		self.conn.fromuserqueue.put(b"\x6F\x01\x55\x03\x07\x02\xEE\x12\x59")

	def _test_io_control_composite_did_dict(self):
		values = {'IAC_pintle': 0x07, 'rpm': 0x1234, 'pedalA': 0x4, 'pedalB' : 0x5,  'EGR_duty': 0x99}
		response = self.udsclient.io_control(control_param=3, did=0x155, values=values, masks=['IAC_pintle', 'pedalA'])	# Short Term Adjustment
		expected_values = {
			'IAC_pintle': 0x07,
			'rpm' 		: 0x2EE,
			'pedalA' 	: 0x1,
			'pedalB' 	: 0x2,
			'EGR_duty' 	: 0x59
		}
		self.assertEqual(response.service_data.did_echo, 0x155)	
		self.assertEqual(response.service_data.control_param_echo, 3)	
		self.assertEqual(response.service_data.decoded_data, expected_values)	

	def test_io_control_composite_did_list(self):
		request = self.conn.touserqueue.get(timeout=0.2)
		self.assertEqual(request, b"\x2F\x01\x55\x03\x07\x12\x34\x45\x99\x00\xA0")
		self.conn.fromuserqueue.put(b"\x6F\x01\x55\x03\x07\x02\xEE\x12\x59")	

	def _test_io_control_composite_did_list(self):
		values = [0x07, 0x1234, 0x4, 0x5, 0x99]
		response = self.udsclient.io_control(control_param=3, did=0x155, values=values, masks=['IAC_pintle', 'pedalA'])	# Short Term Adjustment
		expected_values = {
			'IAC_pintle': 0x07,
			'rpm' 		: 0x2EE,
			'pedalA' 	: 0x1,
			'pedalB' 	: 0x2,
			'EGR_duty' 	: 0x59
		}
		self.assertEqual(response.service_data.did_echo, 0x155)	
		self.assertEqual(response.service_data.control_param_echo, 3)	
		self.assertEqual(response.service_data.decoded_data, expected_values)	

	def test_io_control_non_existent_mask_error(self):
		pass

	def _test_io_control_non_existent_mask_error(self):
		values = [0x07, 0x1234, 0x4, 0x5, 0x99]
		with self.assertRaises(ConfigError):
			self.udsclient.io_control(control_param=3, did=0x155, values=values, masks=['xxxx'])	# mask xxxx does not exist in config
	
	def test_io_control_mask_all_set1(self):
		request = self.conn.touserqueue.get(timeout=0.2)
		self.assertEqual(request, b"\x2F\x01\x55\x03\x07\x12\x34\x45\x99\xFF\xFF")
		self.conn.fromuserqueue.put(b"\x6F\x01\x55\x03\x07\x02\xEE\x12\x59")	

	def _test_io_control_mask_all_set1(self):
		values = [0x07, 0x1234, 0x4, 0x5, 0x99]
		self.udsclient.io_control(control_param=3, did=0x155, values=values, masks=True)	# Short Term Adjustment

	def test_io_control_mask_all_set0(self):
		request = self.conn.touserqueue.get(timeout=0.2)
		self.assertEqual(request, b"\x2F\x01\x55\x03\x07\x12\x34\x45\x99\x00\x00")
		self.conn.fromuserqueue.put(b"\x6F\x01\x55\x03\x07\x02\xEE\x12\x59")	

	def _test_io_control_mask_all_set0(self):
		values = [0x07, 0x1234, 0x4, 0x5, 0x99]
		self.udsclient.io_control(control_param=3, did=0x155, values=values, masks=False)	# Short Term Adjustment

	def test_io_control_no_mask(self):
		request = self.conn.touserqueue.get(timeout=0.2)
		self.assertEqual(request, b"\x2F\x01\x55\x03\x07\x12\x34\x45\x99")
		self.conn.fromuserqueue.put(b"\x6F\x01\x55\x03\x07\x02\xEE\x12\x59")	

	def _test_io_control_no_mask(self):
		values = [0x07, 0x1234, 0x4, 0x5, 0x99]
		self.udsclient.io_control(control_param=3, did=0x155, values=values)	# Short Term Adjustment


	def test_io_control_bad_mask_size(self):
		pass

	def _test_io_control_bad_mask_size(self):
		values = [0x07, 0x1234, 0x4, 0x5, 0x99]
		with self.assertRaises(ValueError):
			self.udsclient.config['input_output'][0x155]['mask_size'] = -1  # Bad value
			self.udsclient.io_control(control_param=3, did=0x155, values=values)	

		with self.assertRaises(ValueError):
			self.udsclient.config['input_output'][0x155]['mask_size'] = None   # Bad value
			self.udsclient.io_control(control_param=3, did=0x155, values=values) 

		with self.assertRaises(ConfigError):
			del self.udsclient.config['input_output'][0x155]['mask_size']
			self.udsclient.io_control(control_param=3, did=0x155, values=values, masks=True) # Set mask, no size defined

	def test_io_control_bad_mask(self):
		pass

	def _test_io_control_bad_mask(self):
		values = [0x07, 0x1234, 0x4, 0x5, 0x99]

		with self.assertRaises(ValueError):
			self.udsclient.config['input_output'][0x155]['mask']['pedalA'] = -1  # Bad value
			self.udsclient.io_control(control_param=3, did=0x155, values=values)	

		with self.assertRaises(ValueError):
			self.udsclient.config['input_output'][0x155]['mask']['pedalA'] = 0x10000  # Bigger than max_size (2)
			self.udsclient.io_control(control_param=3, did=0x155, values=values) 

	def test_io_control_bad_values(self):
		pass

	def _test_io_control_bad_values(self):
		with self.assertRaises(ValueError):
			self.udsclient.io_control(control_param=3, did=0x155, values=1)

		with self.assertRaises(ValueError):
			self.udsclient.io_control(control_param=3, did=0x155, values='asd') 

	def test_io_control_bad_control_param(self):
		pass

	def _test_io_control_bad_control_param(self):
		with self.assertRaises(ValueError):
			self.udsclient.io_control(control_param=-1, did=0x155, values=1)

		with self.assertRaises(ValueError):
			self.udsclient.io_control(control_param=0x100, did=0x155, values='asd') 

	def test_io_control_bad_response_too_much_data_exception(self):
		self.wait_request_and_respond(b"\x6F\x01\x32\x01\x4B\xAA")	# Last byte is extra

	def _test_io_control_bad_response_too_much_data_exception(self):
		with self.assertRaises(InvalidResponseException):
			self.udsclient.io_control(control_param=1, did=0x132)	

	def test_io_control_bad_response_too_much_data_no_exception(self):
		self.wait_request_and_respond(b"\x6F\x01\x32\x01\x4B\xAA")	# Last byte is extra

	def _test_io_control_bad_response_too_much_data_no_exception(self):
		self.udsclient.config['exception_on_invalid_response'] = False
		response = self.udsclient.io_control(control_param=1, did=0x132)
		self.assertFalse(response.valid)

	def test_io_control_bad_response_wrong_control_param_exception(self):
		self.wait_request_and_respond(b"\x6F\x01\x32\x04\x4B")	# 0x04 should be 0x03

	def _test_io_control_bad_response_wrong_control_param_exception(self):
		with self.assertRaises(UnexpectedResponseException):
			self.udsclient.io_control(control_param=3, did=0x132)	

	def test_io_control_bad_response_wrong_control_param_no_exception(self):
		self.wait_request_and_respond(b"\x6F\x01\x32\x04\x4B")	# 0x04 should be 0x03

	def _test_io_control_bad_response_wrong_control_param_no_exception(self):
		self.udsclient.config['exception_on_unexpected_response'] = False
		response = self.udsclient.io_control(control_param=3, did=0x132)
		self.assertTrue(response.valid)
		self.assertTrue(response.unexpected)



from . import *
from udsoncan.Response import Response
from udsoncan.exceptions import *

class SecurityAccess(BaseService):
	_sid = 0x27

	supported_negative_response = [	Response.Code.SubFunctionNotSupported, 
							Response.Code.IncorrectMessageLegthOrInvalidFormat,
							Response.Code.ConditionsNotCorrect,
							Response.Code.RequestSequenceError,
							Response.Code.RequestOutOfRange,
							Response.Code.InvalidKey,
							Response.Code.ExceedNumberOfAttempts,
							Response.Code.RequiredTimeDelayNotExpired
							]

	class Mode:
		RequestSeed=0
		SendKey=1

	@classmethod 
	def normalize_level(cls, mode, level):
		cls.validate_mode(mode)
		ServiceHelper.validate_int(level, min=0, max=0x7F, name='Security level')

		if mode == cls.Mode.RequestSeed:
			return level if level % 2 == 1 else level-1
		elif mode == cls.Mode.SendKey:
			return level if level % 2 == 0 else level+1

	@classmethod
	def make_request(cls, level, mode, key=None):
		"""
		Generates a request for SecurityAccess

		:param level: Service subfunction. The security level to unlock. Value ranging from 0 to 7F 
			For mode=``RequestSeed`` (0), level must be an even value. For mode=``SendKey`` (1), level must be an odd value.
			If the even/odd constraint is not respected, the level value will be corrected to properly set the LSB.
		:type level: int

		:param mode: Type of request to perform. ``SecurityAccess.Mode.RequestSeed`` or ``SecurityAccess.Mode.SendKey`` 
		:type mode: SecurityAccess.Mode, int

		:param key: When mode=``SendKey``, this value must be provided.
		:type key: bytes

		:raises ValueError: If parameters are out of range, missing or wrong type
		"""		
		from udsoncan import Request
		cls.validate_mode(mode)

		ServiceHelper.validate_int(level, min=0, max=0x7F, name='Security level')
		req = Request(service=cls, subfunction=cls.normalize_level(mode=mode, level=level))

		if mode == cls.Mode.SendKey:
			if not isinstance(key, bytes):
				raise ValueError('key must be a valid bytes object')
			req.data = key

		return req

	@classmethod
	def interpret_response(cls, response, mode):
		"""
		Populates the response ``service_data`` property with an instance of :class:`SecurityAccess.ResponseData<udsoncan.services.SecurityAccess.ResponseData>`

		:param response: The received response to interpret
		:type response: :ref:`Response<Response>`

		:raises InvalidResponseException: If length of ``response.data`` is too short
		:raises ValueError: If mode is not ``RequestSeed`` or ``SendKey``
		"""

		cls.validate_mode(mode)

		response.service_data = cls.ResponseData()
		minlength = 2 if mode == cls.Mode.RequestSeed else 1

		if len(response.data) < minlength:
				raise InvalidResponseException(response, "Response data must be at least %d bytes" % (minlength))
		
		response.service_data.security_level_echo = response.data[0]

		if mode == cls.Mode.RequestSeed:
			response.service_data.seed = response.data[1:]

	@classmethod
	def validate_mode(cls, mode):
		if mode not in [cls.Mode.RequestSeed, cls.Mode.SendKey]:
			raise ValueError('Given mode must be either be RequestSeed (0) or SendKey (1).')

	class ResponseData(BaseResponseData):
		"""
		.. data:: security_level_echo

			Requests subfunction echoed back by the server

		.. data:: seed

			Seed value. Only present if request mode was ``RequestSeed`` (even subfunction)
		"""		
		def __init__(self):
			super().__init__(SecurityAccess)

			self.security_level_echo = None
			self.seed = None